"""Dask-enabled Baseband reader classes."""

import contextlib
import dask
import dask.array as da
from .baseband_readers import (BasebandReader, BasebandRawReader,
                               GUPPIRawReader, DADAStokesReader)

__all__ = [
    'DaskBasebandReader', 'DaskBasebandRawReader', 'DaskGUPPIRawReader',
    'DaskDADAStokesReader'
]


class DaskBasebandReader(BasebandReader):
    """Dask-enabled wrapper around StreamReader from the `~baseband` package.

    Parameters
    ----------
    name, **kwargs
        Arguments to pass on to `~baseband.open` to create a StreamReader
        object via `baseband.open(name, 'rs', **kwargs)`.
    """
    @dask.delayed(pure=False)
    def _read_delayed(self, n, offset, /, lock=contextlib.nullcontext()):
        with lock:
            return self._read_array(n, offset)

    def _read_stream(self, n, /, **kwargs):
        """Read N samples from current stream position into array-like."""
        z = da.from_delayed(self._read_delayed(n, self.offset, **kwargs),
                            dtype=self.dtype, shape=(n,) + self.sample_shape)

        _chunk = (-1, ) + ('auto', ) * len(self.sample_shape)
        return da.rechunk(z, chunks=_chunk)


class DaskBasebandRawReader(DaskBasebandReader, BasebandRawReader):
    """Dask-enabled baseband reader for raw voltage data."""
    pass


class DaskGUPPIRawReader(DaskBasebandReader, GUPPIRawReader):
    """Dask-enabled baseband reader for GUPPI raw voltage data format."""
    pass


class DaskDADAStokesReader(DaskBasebandReader, DADAStokesReader):
    """Dask-enabled baseband reader for GUPPI raw voltage data format."""
    pass



"""Core module defining Signal classes."""

import inspect
import numpy as np
import astropy.units as u
from astropy.time import Time

__all__ = [
    'Signal',
    'RadioSignal',
    'IntensitySignal',
    'BasebandSignal',
    'FullStokesSignal',
    'DualPolarizationSignal',
]


class InvalidSignalError(ValueError):
    """Used to catch invalid signals."""
    pass


class Signal:
    """Base class for all signals.

    A signal is sufficiently described by an array of samples (`z`),
    and a sampling rate (`sample_rate`). Optionally, a `start_time` can
    be provided to specify the time stamp at the first sample of the
    signal.

    The signal data (`z`) must have at least 1 dimension where the
    zeroth axis (`axis=0`) refers to time. That is, `z[i]` is the `i`-th
    sample of the signal, and the sample shape (`z[i].shape`) can be
    arbitrary.

    Parameters
    ----------
    z : array
        The signal data. Must be at least 1-dimensional with shape
        `(nsample, ...)`, and must have non-zero size.
    sample_rate : :py:class:`astropy.units.Quantity`
        The number of samples per second. Must be in units of frequency.
    start_time : :py:class:`astropy.time.Time`, optional
        The start time of the signal (that is, the time at the first
        sample of the signal). Default is None.
    """
    _req_dtype = None
    _req_shape = (None, )

    def __init__(self, z, /, *, sample_rate, start_time=None):
        try:
            self._sample_rate = sample_rate.to(u.MHz)
            assert self._sample_rate.isscalar and self._sample_rate > 0
        except Exception:
            err = ("Invalid sample_rate. Must be a positive scalar "
                   "Quantity with units of Hz or equivalent.")
            raise ValueError(err)

        self._start_time = None
        if start_time is not None:
            try:
                self._start_time = Time(start_time, format='isot', precision=9)
                assert self._start_time.isscalar
            except Exception:
                err = ("Invalid start_time. Must be a scalar "
                       "astropy.time.Time object.")
                raise ValueError(err)

        if z.size == 0:
            raise InvalidSignalError("Signal has zero size.")

        min_ndim = len(self._req_shape)

        if z.ndim < min_ndim:
            err = (f"Expected signal with at least {min_ndim} dimension(s), "
                   f"got signal with {z.ndim} dimension(s) instead.")
            raise InvalidSignalError(err)

        zipped = zip(z.shape[:min_ndim], self._req_shape)
        if not all(x == (y or x) for x, y in zipped):
            err = (f"Signal has invalid shape. Expected {self._req_shape}, "
                   f"got {z.shape} instead.")
            raise InvalidSignalError(err)

        if self._req_dtype is not None and z.dtype not in self._req_dtype:
            err = (f"Signal has invalid dtype. Expected {self._req_dtype}, "
                   f"got {z.dtype} instead.")
            raise InvalidSignalError(err)
        self._data = z

    def __str__(self):
        signature = f"{self.__class__.__name__} @ {hex(id(self))}"
        c = type(self.data)
        s = f"{signature}\n{'-' * len(signature)}\n"
        s += f"Data Container: {c.__module__}.{c.__name__}"
        s += f"<shape={self.shape}, dtype={self.dtype}>\n"
        s += f"Sample rate: {self.sample_rate}\n"
        s += f"Time length: {self.time_length}\n"
        st = 'N/A' if self.start_time is None else self.start_time.isot
        s += f"Start time: {st}"
        return s

    def __repr__(self):
        info = f"shape={self.shape}, dtype={self.dtype}"
        s = f"pulsarbat.{self.__class__.__name__}<{info}> @ {hex(id(self))}"
        return s

    def __len__(self):
        return len(self.data)

    def __array__(self):
        return np.asanyarray(self.data)

    def _time_slice(self, index):
        s = slice(*index.indices(self.shape[0]))
        assert s.step > 0, "Time axis slicing does not support negative step"
        assert s.stop > s.start, "Empty time slice!"

        kw = dict()
        if s.step > 1:
            kw['sample_rate'] = self.sample_rate / s.step
        if self.start_time is not None:
            kw['start_time'] = self.start_time + s.start / self.sample_rate
        return kw

    def __getitem__(self, index):
        if not isinstance(index, tuple):
            index = (index,)

        if not all(isinstance(a, slice) for a in index[:1]):
            err = "Only supports slicing on time axis."
            raise IndexError(err)

        kw = dict()
        kw.update(self._time_slice(index[0]))
        return type(self).like(self, self.data[index], **kw)

    @property
    def data(self):
        """The signal data."""
        return self._data

    @property
    def shape(self):
        """Shape of the signal."""
        return self.data.shape

    @property
    def sample_shape(self):
        """Shape of a sample."""
        return self.shape[1:]

    @property
    def ndim(self):
        """Number of dimensions in data."""
        return self.data.ndim

    @property
    def dtype(self):
        """Data type of the signal."""
        return self.data.dtype

    @property
    def sample_rate(self):
        """Sample rate of the signal."""
        return self._sample_rate

    @property
    def dt(self):
        """Sample spacing of the signal (1 / sample_rate)."""
        return (1 / self.sample_rate).to(u.s)

    @property
    def time_length(self):
        """Length of signal in time units."""
        return (len(self) / self.sample_rate).to(u.s)

    @property
    def start_time(self):
        """Start time of the signal (Time at first sample)."""
        return self._start_time

    @property
    def stop_time(self):
        """Stop time of the signal (Time at sample after the last sample)."""
        if self.start_time is None:
            return None
        return self.start_time + self.time_length

    def __contains__(self, time):
        """Tell if `time` is within the bounds of the signal."""
        if self.start_time is None:
            return False
        return self.start_time <= time + 0.1 * u.ns < self.stop_time

    @classmethod
    def like(cls, obj, z=None, /, **kwargs):
        """Creates a signal object "like" another signal object.

        This classmethod inspects the class signature and creates a
        signal object like the given reference object. The signal data
        (as an array) must always be provided as well as any other
        arguments that are intended to be different than the reference
        object. All arguments required by the class signature that are
        not provided are acquired from attributes of the same name in
        the reference object.

        Parameters
        ----------
        obj : `Signal`
            The reference signal object.
        z : `~numpy.ndarray`-like
            The signal data.
        **kwargs
            Additional keyword arguments to pass on to the target class.
        """
        sig = inspect.signature(cls)

        for k, v in sig.parameters.items():
            if v.kind is not v.POSITIONAL_ONLY and k not in kwargs:
                if hasattr(obj, k):
                    kwargs[k] = getattr(obj, k)
                elif v.default is v.empty:
                    raise ValueError(f'Missing required keyword argument: {k}')

        if z is None:
            z = obj.data
        return cls(z, **kwargs)


class RadioSignal(Signal):
    """Class for heterodyned radio signals.

    A radio signal is often heterodyned, so a center frequency
    (`center_freq`) and a channel bandwidth (`chan_bw`) must be provided
    to determine the frequencies of each channel. Optionally,
    `freq_align` is provided to indicate how the channel frequencies are
    positioned. Usually, the signal data here would be the result of
    some sort of filterbank.

    The signal data (`z`) must have at least 2 dimensions where the
    first two dimensions refer to time (`axis=0`) and frequency
    (`axis=1`), respectively.

    Parameters
    ----------
    z : `~numpy.ndarray`-like
        The signal data. Must be at least 2-dimensional with shape
        `(nsample, nchan, ...)`.
    sample_rate : `~astropy.units.Quantity`
        The number of samples per second. Must be in units of frequency.
    start_time : `~astropy.time.Time`, optional
        The start time of the signal (that is, the time at the first
        sample of the signal). Default is None.
    center_freq : `~astropy.units.Quantity`
        The frequency at the center of the signal's band. Must be in
        units of frequency.
    chan_bw : `~astropy.units.Quantity`
        The bandwidth of a channel. The total bandwidth is `chan_bw *
        nchan`. Must be in units of frequency.
    freq_align : {'bottom', 'center', 'top'}, optional
        The alignment of channels relative to the `center_freq`. Default
        is `'center'` (as with odd-length complex DFTs). `'bottom'` and
        `'top'` only have an effect when `nchan` is even.

    Notes
    -----
    The input signal array must have shape `(nsample, nchan, ...)`,
    where `nsample` is the number of time samples, and `nchan` is the
    number of frequency channels.

    The channels must be adjacent in frequency and of equal channel
    bandwidth, such that the frequency of a channel `i` is given by,::

        freq_i = center_freq + chan_bw * (i + a - nchan/2)

    where `i` is in `[0, ..., nchan - 1]`, `nchan` is the number of
    channels (`z.shape[1]`), `chan_bw` is the bandwidth of a single
    channel, and `a` is in `{0, 0.5, 1}` depending on the value of
    `freq_align`.

    An even-length complex-valued DFT (as implemented in `~numpy.fft`)
    would have `freq_align = 'bottom'` with a center frequency of 0,
    whereas an odd-length DFT would have `freq_align = 'center'` due to
    symmetry. `freq_align = 'top'` is used in cases where the DFT was
    conducted on the lower sideband of the signal instead.

    When `nchan` is odd, the `freq_align` attribute will internally be
    fixed to `'center'` since the channels must be necessarily centered
    on the `center_freq`.
    """
    _req_shape = (None, None, )

    def __init__(self, z, /, *, sample_rate, start_time=None, center_freq,
                 chan_bw, freq_align='center'):

        try:
            self._center_freq = center_freq.to(u.MHz)
            assert self._center_freq.isscalar
        except Exception:
            err = ("Invalid center_freq. Must be a scalar "
                   "Quantity with units of Hz or equivalent.")
            raise ValueError(err)

        try:
            self._chan_bw = chan_bw.to(u.MHz)
            assert self._chan_bw.isscalar and self._chan_bw > 0
        except Exception:
            err = ("Invalid chan_bw. Must be a positive scalar "
                   "Quantity with units of Hz or equivalent.")
            raise ValueError(err)

        super().__init__(z, sample_rate=sample_rate, start_time=start_time)

        if freq_align in ['bottom', 'center', 'top']:
            if self.nchan % 2:
                self._freq_align = 'center'
            else:
                self._freq_align = freq_align
        else:
            choices = "{'bottom', 'center', 'top'}"
            raise ValueError(f'Invalid freq_align. Expected: {choices}')

    def __str__(self):
        s = super().__str__() + "\n"
        s += f"Channel Bandwidth: {self.chan_bw}\n"
        s += f"Total Bandwidth: {self.bandwidth}\n"
        s += f"Center Frequency: {self.center_freq}"
        return s

    def _freq_slice(self, index):
        s = slice(*index.indices(self.shape[1]))
        assert s.step == 1, "Does not support slice step for frequency axis"
        assert s.stop > s.start, "Empty frequency slice!"
        kw = dict()
        if s.stop - s.start < self.nchan:
            f = self.channel_freqs[s]
            kw['center_freq'] = (f[0] + f[-1]) / 2
            kw['freq_align'] = 'center'
        return kw

    def __getitem__(self, index):
        if not isinstance(index, tuple):
            index = (index,)

        if not all(isinstance(a, slice) for a in index[:2]):
            err = "Only supports slicing on time and frequency axes."
            raise IndexError(err)

        kw = dict()
        kw.update(self._time_slice(index[0]))
        if len(index) > 1:
            kw.update(self._freq_slice(index[1]))
        return type(self).like(self, self.data[index], **kw)

    @property
    def nchan(self):
        """Number of frequency channels."""
        return self.shape[1]

    @property
    def center_freq(self):
        """Center frequency."""
        return self._center_freq

    @property
    def bandwidth(self):
        """Total bandwidth."""
        return self.chan_bw * self.nchan

    @property
    def chan_bw(self):
        """Channel bandwidth."""
        return self._chan_bw

    @property
    def max_freq(self):
        """Frequency at top of the band."""
        return self.center_freq + self.bandwidth / 2

    @property
    def min_freq(self):
        """Frequency at bottom of the band."""
        return self.center_freq - self.bandwidth / 2

    @property
    def freq_align(self):
        """Alignment of channel frequencies."""
        return self._freq_align

    @property
    def channel_freqs(self):
        """Returns a list of frequencies corresponding to all channels."""
        _align = {'bottom': 0, 'center': 0.5, 'top': 1}[self.freq_align]
        chan_ids = np.arange(self.nchan) + _align - self.nchan / 2
        return self.center_freq + self.chan_bw * chan_ids


class IntensitySignal(RadioSignal):
    """Class for intensity signals such as Stokes I, Q, U, V, etc.

    See the documentation for `~RadioSignal` for more details.

    Parameters
    ----------
    z : `~numpy.ndarray`-like
        The signal data. Must be at least 2-dimensional with shape
        `(nsample, nchan, ...)`, and dtype of either `np.float32` or
        `np.float64`.
    sample_rate : `~astropy.units.Quantity`
        The number of samples per second. Must be in units of frequency.
    start_time : `~astropy.time.Time`, optional
        The start time of the signal (that is, the time at the first
        sample of the signal). Default is None.
    center_freq : `~astropy.units.Quantity`
        The frequency at the center of the signal's band. Must be in
        units of frequency.
    chan_bw : `~astropy.units.Quantity`
        The bandwidth of a channel. The total bandwidth is `chan_bw *
        nchan`. Must be in units of frequency.
    freq_align : {'bottom', 'center', 'top'}, optional
        The alignment of channels relative to the `center_freq`. Default
        is `'center'` (as with odd-length complex DFTs). `'bottom'` and
        `'top'` only have an effect when `nchan` is even.

    See Also
    --------
    Signal
    RadioSignal
    """
    _req_dtype = (np.float32, np.float64)


class FullStokesSignal(IntensitySignal):
    """Class for full Stokes (I, Q, U, V) signals.

    See the documentation for `~RadioSignal` and `~IntensitySignal` for
    more details.

    Parameters
    ----------
    z : `~numpy.ndarray`-like
        The signal data. Must be at least 3-dimensional with shape
        `(nsample, nchan, nstokes, ...)` where `nstokes = 4`, and dtype
        of either `np.float32` or `np.float64`. The order of the Stokes
        components is `[I, Q, U, V]`.
    sample_rate : `~astropy.units.Quantity`
        The number of samples per second. Must be in units of frequency.
    start_time : `~astropy.time.Time`, optional
        The start time of the signal (that is, the time at the first
        sample of the signal). Default is None.
    center_freq : `~astropy.units.Quantity`
        The frequency at the center of the signal's band. Must be in
        units of frequency.
    chan_bw : `~astropy.units.Quantity`
        The bandwidth of a channel. The total bandwidth is `chan_bw *
        nchan`. Must be in units of frequency.
    freq_align : {'bottom', 'center', 'top'}, optional
        The alignment of channels relative to the `center_freq`. Default
        is `'center'` (as with odd-length complex DFTs). `'bottom'` and
        `'top'` only have an effect when `nchan` is even.

    See Also
    --------
    Signal
    RadioSignal
    IntensitySignal

    References
    ----------
    Wikipedia, "Stokes Parameters",
    https://en.wikipedia.org/wiki/Stokes_parameters
    """
    _req_shape = (None, None, 4)


class BasebandSignal(RadioSignal):
    """Class for complex baseband signals.

    Baseband signals are Nyquist-sampled analytic signals. In the
    complex baseband representation, the signal is shifted to be
    centered around zero frequency and is a complex-valued signal. As a
    consequence, `sample_rate == chan_bw`. All frequency channels are
    assumed to be upper side-band (that is, the signal isn't spectrally
    flipped).

    See the documentation for `~RadioSignal` for more details.

    Parameters
    ----------
    z : `~numpy.ndarray`-like
        The signal data. Must be at least 2-dimensional with shape
        `(nsample, nchan, ...)`, and dtype of either `np.complex64` or
        `np.complex128`.
    sample_rate : `~astropy.units.Quantity`
        The number of samples per second. Must be in units of frequency.
    start_time : `~astropy.time.Time`, optional
        The start time of the signal (that is, the time at the first
        sample of the signal). Default is None.
    center_freq : `~astropy.units.Quantity`
        The frequency at the center of the signal's band. Must be in
        units of frequency.
    freq_align : {'bottom', 'center', 'top'}, optional
        The alignment of channels relative to the `center_freq`. Default
        is `'center'` (as with odd-length complex DFTs). `'bottom'` and
        `'top'` only have an effect when `nchan` is even.

    See Also
    --------
    Signal
    RadioSignal
    """
    _req_dtype = (np.complex64, np.complex128)

    def __init__(self, z, /, *, sample_rate, start_time=None, center_freq,
                 freq_align='center'):

        super().__init__(z, sample_rate=sample_rate, start_time=start_time,
                         center_freq=center_freq, chan_bw=sample_rate,
                         freq_align=freq_align)

    def to_intensity(self):
        """Converts baseband signal to intensities.

        Returns
        -------
        out : `~pulsarbat.IntensitySignal`
            The converted signal.
        """
        z = self.data.real**2 + self.data.imag**2
        return IntensitySignal.like(self, z)


class DualPolarizationSignal(BasebandSignal):
    """Class for dual-polarization complex baseband signals.

    See the documentation for `~RadioSignal` and `~BasebandSignal` for
    more details.

    Parameters
    ----------
    z : `~numpy.ndarray`-like
        The signal data. Must be at least 3-dimensional with shape
        `(nsample, nchan, npol, ...)` where `npol = 2`, and dtype of
        either `np.complex64` or `np.complex128`.
    sample_rate : `~astropy.units.Quantity`
        The number of samples per second. Must be in units of frequency.
    start_time : `~astropy.time.Time`, optional
        The start time of the signal (that is, the time at the first
        sample of the signal). Default is None.
    center_freq : `~astropy.units.Quantity`
        The frequency at the center of the signal's band. Must be in
        units of frequency.
    freq_align : {'bottom', 'center', 'top'}, optional
        The alignment of channels relative to the `center_freq`. Default
        is `'center'` (as with odd-length complex DFTs). `'bottom'` and
        `'top'` only have an effect when `nchan` is even.
    pol_type : {'linear', 'circular'}
        The polarization type of the signal. `'linear'` for linearly
        polarized signals (with basis `[X, Y]`) or `'circular'` for
        circularly polarized signals (with basis `[R, L]`).

    See Also
    --------
    Signal
    RadioSignal
    BasebandSignal

    Notes
    -----
    Linear polarization is assumed to have basis `[X, Y]` and circular
    polarization is assumed to have basis `[R, L]`. For example, with
    `pol_type='circular'`, `z[:, :, 0]` refers to the right-handed
    circular polarization component.
    """
    _req_shape = (None, None, 2)

    def __init__(self, z, /, *, sample_rate, start_time=None,
                 center_freq, freq_align='center', pol_type):

        if pol_type in ['linear', 'circular']:
            self._pol_type = pol_type
        else:
            raise ValueError("pol_type must be in {'linear', 'circular'}")

        super().__init__(z, sample_rate=sample_rate, start_time=start_time,
                         center_freq=center_freq, freq_align=freq_align)

    def __str__(self):
        s = super().__str__() + "\n"
        s += f"Polarization Type: {self.pol_type}"
        return s

    @property
    def pol_type(self):
        return self._pol_type

    def to_linear(self):
        """Converts the dual-polarization signal to linear basis.

        If polarization basis is already linear, simply returns a copy of
        the signal object.

        Returns
        -------
        out : `~pulsarbat.DualPolarizationSignal`
            The converted signal.
        """
        if self.pol_type == 'circular':
            R, L = self.data[:, :, 0], self.data[:, :, 1]
            X, Y = R + L, 1j * (R - L)
            z = np.stack([X, Y], axis=2) / np.sqrt(2)
        else:
            z = self.data

        return type(self).like(self, z, pol_type='linear')

    def to_circular(self):
        """Converts the dual-polarization signal to circular basis.

        If polarization basis is already circular, simply returns a copy of
        the signal object.

        Returns
        -------
        out : `~pulsarbat.DualPolarizationSignal`
            The converted signal.
        """
        if self.pol_type == 'linear':
            X, Y = self.data[:, :, 0], self.data[:, :, 1]
            R, L = X - 1j * Y, X + 1j * Y
            z = np.stack([R, L], axis=2) / np.sqrt(2)
        else:
            z = self.data

        return type(self).like(self, z, pol_type='circular')

    def to_stokes(self):
        """Converts signal to IQUV Stokes representation.

        Returns
        -------
        out : `~pulsarbat.StokesSignal`
            Signal in Stokes IQUV representation.
        """
        if self.pol_type == 'linear':
            X, Y = self.data[:, :, 0], self.data[:, :, 1]

            i = X.real**2 + X.imag**2 + Y.real**2 + Y.imag**2
            Q = X.real**2 + X.imag**2 - Y.real**2 - Y.imag**2
            U = +2 * (X * Y.conj()).real
            V = -2 * (X * Y.conj()).imag

        elif self.pol_type == 'circular':
            R, L = self.data[:, :, 0], self.data[:, :, 1]

            i = R.real**2 + R.imag**2 + L.real**2 + L.imag**2
            Q = +2 * (R * L.conj()).real
            U = -2 * (R * L.conj()).imag
            V = R.real**2 + R.imag**2 - L.real**2 - L.imag**2

        z = np.stack([i, Q, U, V], axis=2)
        return FullStokesSignal.like(self, z)



"""Top-level module for pulsarbat."""

# flake8: noqa

__author__ = """Nikhil Mahajan"""
__email__ = 'mahajan@astro.utoronto.ca'
__version__ = '0.0.4'


from . import core
from . import signal_funcs
from . import reader
from . import utils

from .core import *
from .signal_funcs import *

__all__ = []
__all__.extend(core.__all__)
__all__.extend(signal_funcs.__all__)



"""Tests for `pulsarbat.RadioSignal` and subclasses."""

import pytest
import numpy as np
import pulsarbat as pb


@pytest.mark.parametrize("in_type, out_type", [(np.float32, np.complex64),
                                               (np.float64, np.complex128),
                                               (np.complex64, np.complex64),
                                               (np.complex128, np.complex128)])
@pytest.mark.parametrize("N", [128, 127])
def test_real_to_complex(in_type, out_type, N):
    x = np.random.normal(0, 1, N).astype(in_type)
    y = pb.utils.real_to_complex(x, axis=0)
    assert y.dtype == out_type
    assert y.shape[0] == (N + 1) // 2



"""Tests for `pulsarbat.RadioSignal` and subclasses."""

import pytest
import numpy as np
import astropy.units as u
import pulsarbat as pb
from astropy.time import Time


@pytest.mark.parametrize("start_time", [None, Time.now()])
def test_concatenate_time(start_time):
    shape = (4096, 32, 2)
    SR = (100 * u.MHz) / shape[1]
    x = np.random.normal(0, 1, shape) + np.random.normal(0, 1, shape) * 1j
    z = pb.DualPolarizationSignal(x, sample_rate=SR, center_freq=400*u.MHz,
                                  pol_type='linear', start_time=start_time)

    y = pb.concatenate([z], axis=0)
    assert np.allclose(np.array(z), np.array(y))
    for a in ('center_freq', 'chan_bw', 'sample_rate'):
        assert u.isclose(getattr(z, a), getattr(y, a))

    if start_time is None:
        assert y.start_time is None
    else:
        assert abs(y.start_time - z.start_time) < 0.1 * u.ns

    idx = [0, 16, 128, 512, 2000, 3000, 4096]
    s = [z[i:j] for i, j in zip(idx, idx[1:])]
    y = pb.concatenate(s, axis=0)
    assert np.allclose(np.array(z), np.array(y))
    for a in ('center_freq', 'chan_bw', 'sample_rate'):
        assert u.isclose(getattr(z, a), getattr(y, a))

    if start_time is None:
        assert y.start_time is None
    else:
        assert abs(y.start_time - z.start_time) < 0.1 * u.ns


@pytest.mark.parametrize("freq_align", ['bottom', 'center', 'top'])
def test_concatenate_freq(freq_align):
    shape = (16, 2048, 4)
    fcen = 7*u.GHz
    chan_bw = (2 * u.GHz) / shape[1]
    x = np.random.normal(0, 1, shape)
    z = pb.FullStokesSignal(x, sample_rate=1*u.Hz, chan_bw=chan_bw,
                            center_freq=fcen, freq_align=freq_align)

    cen = {'bottom': fcen - chan_bw / 2,
           'center': fcen,
           'top': fcen + chan_bw / 2}

    y = pb.concatenate([z], axis=1)
    assert np.allclose(np.array(z), np.array(y))
    assert y.freq_align == 'center'
    for a in ('chan_bw', 'sample_rate'):
        assert u.isclose(getattr(z, a), getattr(y, a))
    assert u.allclose(z.channel_freqs, y.channel_freqs)
    assert u.isclose(y.center_freq, cen[freq_align])

    idx = [0, 2, 128, 129, 512, 1234, 1999, 2048]
    s = [z[:, i:j] for i, j in zip(idx, idx[1:])]
    y = pb.concatenate(s, axis=1)
    assert y.freq_align == 'center'
    for a in ('chan_bw', 'sample_rate'):
        assert u.isclose(getattr(z, a), getattr(y, a))
    assert u.allclose(z.channel_freqs, y.channel_freqs)
    assert u.isclose(y.center_freq, cen[freq_align])


def test_concatenate_errors():
    x = np.random.normal(0, 1, (32, 32, 4))
    z = pb.FullStokesSignal(x, sample_rate=1*u.Hz, chan_bw=1*u.MHz,
                            center_freq=1*u.GHz, freq_align='center',
                            start_time=Time.now())

    with pytest.raises(TypeError):
        _ = pb.concatenate(["!", z])

    with pytest.raises(TypeError):
        _ = pb.concatenate([z, "!"])

    with pytest.raises(TypeError):
        _ = pb.concatenate([pb.Signal.like(z)], axis=1)

    with pytest.raises(ValueError):
        _ = pb.concatenate([z], axis=2)

    with pytest.raises(ValueError):
        _ = pb.concatenate([z, z], axis=0)

    with pytest.raises(ValueError):
        _ = pb.concatenate([z, z], axis=1)

    with pytest.raises(ValueError):
        y = pb.FullStokesSignal.like(z, sample_rate=2*u.Hz)
        _ = pb.concatenate([z, y], axis=0)

    with pytest.raises(ValueError):
        y = pb.FullStokesSignal.like(z, chan_bw=z.chan_bw/2)
        _ = pb.concatenate([z, y], axis=1)



"""Tests for `pulsarbat.RadioSignal` and subclasses."""

import pytest
import numpy as np
import astropy.units as u
import pulsarbat as pb
from astropy.time import Time


@pytest.mark.parametrize("ref_freq", [3.95, 4, 4.05] * u.GHz)
@pytest.mark.parametrize("dm", [1, 10])
def test_coherent_dedispersion_reversibility(ref_freq, dm):
    shape = (8192, 4)
    x = np.random.normal(0, 1, shape) + np.random.normal(0, 1, shape) * 1j
    DM = pb.DispersionMeasure(dm)
    z = pb.BasebandSignal(x, sample_rate=16*u.MHz, center_freq=4*u.GHz)

    y = pb.coherent_dedispersion(z, DM, ref_freq=ref_freq)
    y = pb.coherent_dedispersion(y, -DM, ref_freq=ref_freq)


@pytest.mark.parametrize("ref_freq", [3, 5, 7] * u.GHz)
@pytest.mark.parametrize("dm", [100, -100])
def test_incoherent_dedispersion_reversibility(ref_freq, dm):
    kw = {'sample_rate': 10 * u.kHz,
          'center_freq': 5 * u.GHz,
          'chan_bw': 10 * u.MHz,
          'start_time': Time.now()}

    x = pb.RadioSignal(np.random.standard_normal((4096, 32)), **kw)
    DM = pb.DispersionMeasure(dm)

    y = pb.incoherent_dedispersion(x, DM, ref_freq=ref_freq)
    y = pb.incoherent_dedispersion(y, -DM, ref_freq=ref_freq)

    i = int(((y.start_time - x.start_time) / x.dt).to_value(u.one).round())
    assert np.allclose(x.data[i:i + len(y)], y)



"""Baseband reader classes."""

import baseband
import numpy as np
import astropy.units as u
from astropy.time import Time

from .base import AbstractReader
from ..core import (BasebandSignal, DualPolarizationSignal,
                    FullStokesSignal)
from ..utils import real_to_complex

__all__ = ['BasebandReader', 'BasebandRawReader', 'GUPPIRawReader',
           'DADAStokesReader']


class BasebandReader(AbstractReader):
    """Wrapper around StreamReader from the `~baseband` package.

    Parameters
    ----------
    name, **kwargs
        Arguments to pass on to `~baseband.open` to create a StreamReader
        object via `baseband.open(name, 'rs', **kwargs)`.
    """
    def __init__(self, name, /, **kwargs):
        super().__init__()
        self._name = name
        self._kwargs = kwargs
        with self._get_fh() as fh:
            self._info = fh.info
            self._dtype = fh.dtype

    def _get_fh(self):
        return baseband.open(self._name, 'rs', **self._kwargs)

    def __len__(self):
        return self._info.shape[0]

    @property
    def sample_shape(self):
        return self._info.shape[1:]

    @property
    def dtype(self):
        return self._dtype

    @property
    def sample_rate(self):
        """Sample rate of the data."""
        return self._info.sample_rate

    @property
    def start_time(self):
        """Time at first sample of data."""
        return Time(self._info.start_time, format='isot', precision=9)

    def _read_array(self, n, offset):
        """Read n samples from current stream position into numpy array."""
        with self._get_fh() as fh:
            fh.seek(offset)
            z = fh.read(n)
        return z.astype(self.dtype, copy=False)

    def _read_stream(self, n, /, **kwargs):
        """Read N samples from current stream position into array-like."""
        return self._read_array(n, self.offset)


class BasebandRawReader(BasebandReader):
    """Base class for raw voltage data readable by the `~baseband` package.

    Parameters
    ----------
    name, **kwargs
        Arguments to pass on to `~baseband.open` to create a StreamReader
        object via `baseband.open(name, 'rs', **kwargs)`.
    center_freq : `~astropy.units.Quantity`
        The frequency at the center of the signal's band. Must be in
        units of frequency.
    chan_bw : `~astropy.units.Quantity`
        The bandwidth of a channel. The total bandwidth is `chan_bw *
        nchan`. Must be in units of frequency.
    freq_align : {'bottom', 'center', 'top'}, optional
        The alignment of channels relative to the `center_freq`. Default
        is `'center'` (as with odd-length complex DFTs). `'bottom'` and
        `'top'` only have an effect when `nchan` is even.
    sideband : bool or array-like
        Can be True (no spectral flip) or False (spectrally flipped). If
        channels have different sidebands, an array-like object of
        boolean elements can be passed (must have the same shape as the
        StreamReader's `sample_shape`).
    """
    def __init__(self, name, /, *, center_freq, freq_align='center',
                 sideband=True, **kwargs):
        super().__init__(name, **kwargs)

        try:
            self._center_freq = center_freq.to(u.MHz)
            assert self._center_freq.isscalar
        except Exception:
            err = ("Invalid center_freq. Must be a scalar "
                   "Quantity with units of Hz or equivalent.")
            raise ValueError(err)

        if type(sideband) is bool:
            self._sideband = sideband
        else:
            self._sideband = np.array(sideband).astype(bool)
            if self.sample_shape != self.sideband.shape:
                err = "StreamReader sample shape != sideband shape"
                raise ValueError(err)

        if freq_align in ['bottom', 'center', 'top']:
            self._freq_align = freq_align
        else:
            choices = "{'bottom', 'center', 'top'}"
            raise ValueError(f'Invalid freq_align. Expected: {choices}')

    @property
    def dtype(self):
        return np.complex64

    @property
    def sideband(self):
        """True if upper sideband, False if lower sideband."""
        return self._sideband

    @property
    def center_freq(self):
        """Center frequency."""
        return self._center_freq

    @property
    def freq_align(self):
        """Alignment of channel frequencies."""
        return self._freq_align

    @property
    def sample_rate(self):
        """Sample rate (number of samples per unit time)."""
        if self._info.complex_data:
            return self._info.sample_rate
        else:
            return self._info.sample_rate / 2

    def __len__(self):
        if self._info.complex_data:
            return self._info.shape[0]
        else:
            return self._info.shape[0] // 2

    def _read_array(self, n, offset):
        """Read n samples from current stream position into numpy array."""
        with self._get_fh() as fh:
            if self._info.complex_data:
                fh.seek(offset)
                z = fh.read(n)
            else:
                fh.seek(offset * 2)
                z = real_to_complex(fh.read(2*n), axis=0)

        if self.sideband is False:
            z = z.conj()
        elif self.sideband is not True:
            slc = ~self.sideband
            z[:, slc] = z[:, slc].conj()

        return z.astype(self.dtype, copy=False)

    def _to_signal(self, z, /, start_time):
        """Return Signal containing given data."""
        kwargs = {'sample_rate': self.sample_rate,
                  'center_freq': self.center_freq,
                  'freq_align': self.freq_align}
        return BasebandSignal(z, start_time=start_time, **kwargs)


class GUPPIRawReader(BasebandRawReader):
    """Baseband reader for GUPPI raw voltage data format.

    Parameters
    ----------
    name
        File name, filehandle, or sequence of file names to pass on to
        `~baseband.open` to create a GUPPIStreamReader object via
        `baseband.open(name, 'rs', format='guppi')`.
    """

    def __init__(self, name, /):
        kwargs = {'format': 'guppi'}

        with baseband.open(name, 'rs', **kwargs) as fh:
            header = fh.header0

        pol_dict = {'LIN': 'linear', 'CIRC': 'circular'}
        self._pol_type = pol_dict[header['FD_POLN']]

        super().__init__(name, sideband=header.sideband, freq_align='center',
                         center_freq=header['OBSFREQ'] * u.MHz, **kwargs)

    @property
    def pol_type(self):
        return self._pol_type

    def _to_signal(self, z, /, start_time):
        """Return Signal containing given data."""
        kwargs = {'sample_rate': self.sample_rate,
                  'center_freq': self.center_freq,
                  'freq_align': self.freq_align,
                  'pol_type': self.pol_type,
                  'start_time': start_time}
        return DualPolarizationSignal(z.transpose(0, 2, 1), **kwargs)


class DADAStokesReader(BasebandReader):
    """Reader for full Stokes data in DADA format.

    Parameters
    ----------
    name
        File name, filehandle, or sequence of file names to pass on to
        `~baseband.open` to create a DADAStreamReader object via
        `baseband.open(name, 'rs', format='dada')`.
    """
    def __init__(self, name, /):
        kwargs = {'format': 'dada'}
        super().__init__(name, **kwargs)

        with self._get_fh() as fh:
            if not (fh.header0['NPOL'] == 4 and fh.header0['NDIM'] == 1):
                raise ValueError('Does not look like Stokes data')
            self._header = fh.header0

    @property
    def sideband(self):
        return self._header.sideband

    @property
    def center_freq(self):
        return self._header['FREQ'] * u.MHz

    @property
    def chan_bw(self):
        return abs(self._header['BW'] / self._header['NCHAN']) * u.MHz

    @property
    def freq_align(self):
        return 'bottom' if self.sideband else 'top'

    def _to_signal(self, z, /, start_time):
        """Return Signal containing given data."""
        kwargs = {'sample_rate': self.sample_rate,
                  'center_freq': self.center_freq,
                  'chan_bw': self.chan_bw,
                  'freq_align': self.freq_align,
                  'start_time': start_time}

        z = z.transpose(0, 2, 1)
        if not self.sideband:
            z = np.flip(z, axis=1)

        return FullStokesSignal(z, **kwargs)



"""Tests for `pulsarbat.RadioSignal` and subclasses."""

import pytest
import numpy as np
import dask.array as da
import astropy.units as u
from astropy.time import Time
import pulsarbat as pb


@pytest.mark.parametrize("fn", [np.ones, da.ones])
@pytest.mark.parametrize("fcen", [400 * u.MHz, 800 * u.MHz])
@pytest.mark.parametrize("chan_bw", [1 * u.MHz, 10 * u.MHz])
@pytest.mark.parametrize("nchan", [7, 8, 9])
@pytest.mark.parametrize("align", ["bottom", "top", "center"])
def test_radiosignal(fn, fcen, chan_bw, nchan, align):
    z = pb.RadioSignal(fn((16, nchan, 2)), sample_rate=1*u.Hz,
                       center_freq=fcen, chan_bw=chan_bw, freq_align=align)

    bw = chan_bw * nchan
    f = np.fft.fftshift(np.fft.fftfreq(nchan, 1/bw))
    if align == 'center' and nchan % 2 == 0:
        chan_cens = fcen + f + (bw/nchan/2)
    elif align == 'top' and nchan % 2 == 0:
        chan_cens = (fcen - f)[::-1]
    else:
        chan_cens = fcen + f

    assert isinstance(z, pb.RadioSignal)
    print(z)
    repr(z)

    assert z.nchan == nchan
    assert u.isclose(z.center_freq, fcen)
    assert u.isclose(z.bandwidth, bw)
    assert u.isclose(z.chan_bw, chan_bw)
    assert u.isclose(z.max_freq, fcen + bw/2)
    assert u.isclose(z.min_freq, fcen - bw/2)
    print(z.channel_freqs)
    print()
    print(chan_cens, align)
    assert all(u.isclose(z.channel_freqs, chan_cens))

    if nchan % 2:
        assert z.freq_align == 'center'
    else:
        assert z.freq_align == align


def test_shape_errors():
    for shape in [(10,), (24, 0), (4, 4, 0), (0, 5, 10)]:
        with pytest.raises(ValueError):
            _ = pb.RadioSignal(np.empty(shape), sample_rate=1*u.Hz,
                               center_freq=1*u.GHz, chan_bw=1*u.MHz)


def test_center_freq_errors():
    for x in [99., 4*u.m, [5, 6]*u.Hz]:
        with pytest.raises(ValueError):
            _ = pb.RadioSignal(np.empty((8, 4)), sample_rate=1*u.Hz,
                               center_freq=x, chan_bw=10*u.MHz)


def test_chan_bw_errors():
    for x in [99., 4*u.m, [5, 6]*u.Hz, -1*u.Hz]:
        with pytest.raises(ValueError):
            _ = pb.RadioSignal(np.empty((8, 4)), sample_rate=1*u.Hz,
                               center_freq=400*u.MHz, chan_bw=x)


def test_freq_align_errors():
    for x in [None, (), 'invalid']:
        with pytest.raises(ValueError):
            _ = pb.RadioSignal(np.empty((8, 4)), sample_rate=1*u.Hz,
                               center_freq=400*u.MHz, chan_bw=10*u.MHz,
                               freq_align=x)

    for x in ['top', 'center', 'bottom']:
        _ = pb.RadioSignal(np.empty((8, 4)), sample_rate=1*u.Hz,
                           center_freq=400*u.MHz, chan_bw=10*u.MHz,
                           freq_align=x)


@pytest.mark.parametrize("sr", [1*u.MHz, 5*u.MHz])
@pytest.mark.parametrize("nchan", [7, 8])
def test_basebandsignal(sr, nchan):
    x = np.ones((16, nchan)).astype(np.complex64)
    center_freq = 800 * u.MHz

    z = pb.BasebandSignal(x, sample_rate=sr, center_freq=center_freq)
    assert u.isclose(z.sample_rate, sr)
    assert u.isclose(z.chan_bw, sr)
    assert u.isclose(z.bandwidth, sr*nchan)


def test_baseband_dtype():
    fcen, sr = 800*u.MHz, 10*u.MHz

    for dtype in [np.float32, np.float64]:
        x = np.ones((16, 8), dtype=dtype)
        with pytest.raises(ValueError):
            _ = pb.BasebandSignal(x, sample_rate=sr, center_freq=fcen)

    for dtype in [np.complex64, np.complex128]:
        x = np.ones((16, 8), dtype=dtype)
        _ = pb.BasebandSignal(x, sample_rate=sr, center_freq=fcen)


def test_intensity_dtype():
    fcen, sr, cbw = 800*u.MHz, 1*u.Hz, 10*u.MHz

    for dtype in [np.complex64, np.complex128]:
        x = np.ones((16, 8), dtype=dtype)
        with pytest.raises(ValueError):
            _ = pb.IntensitySignal(x, sample_rate=sr, center_freq=fcen,
                                   chan_bw=cbw)

    for dtype in [np.float32, np.float64]:
        x = np.ones((16, 8), dtype=dtype)
        _ = pb.IntensitySignal(x, sample_rate=sr, center_freq=fcen,
                               chan_bw=cbw)


@pytest.mark.parametrize("use_dask", [True, False])
@pytest.mark.parametrize("A", [1, 2, 4, 10])
@pytest.mark.parametrize("in_dtype, out_dtype", [(np.complex64, np.float32),
                                                 (np.complex128, np.float64)])
def test_baseband_to_intensity(A, use_dask, in_dtype, out_dtype):
    shape = (1024, 8, 2)

    if use_dask:
        x = da.random.uniform(-np.pi, np.pi, shape)
    else:
        x = np.random.uniform(-np.pi, np.pi, shape)
    z_phasor = np.exp(1j * x).astype(in_dtype)

    st = Time.now()
    z = pb.BasebandSignal(A * z_phasor, sample_rate=1*u.MHz, start_time=st,
                          center_freq=600 * u.MHz)

    assert isinstance(z.data, type(x))

    zi = z.to_intensity()
    assert isinstance(zi.data, type(x))
    assert np.allclose(A**2, np.array(zi))
    assert abs(st - zi.start_time) < 0.1 * u.ns
    assert u.isclose(z.time_length, zi.time_length)
    assert u.isclose(z.sample_rate, zi.sample_rate)
    assert u.isclose(z.chan_bw, zi.chan_bw)
    assert u.isclose(z.center_freq, zi.center_freq)
    assert z.freq_align == zi.freq_align == 'center'
    assert zi.dtype == out_dtype


def test_radiosignal_slice():
    x = np.ones((16, 16, 4, 2), dtype=np.float32)
    z = pb.RadioSignal(x, sample_rate=1*u.Hz, start_time=Time.now(),
                       center_freq=400*u.MHz, chan_bw=10*u.MHz,
                       freq_align='bottom')

    y = z[2:8, 2:8, 0]
    assert len(y) == 6
    assert u.isclose(y.time_length, 6 * u.s)
    assert np.abs(z.start_time - (y.start_time - 2*u.s)) < 0.1 * u.ns
    assert y.nchan == 6
    assert all(u.isclose(y.channel_freqs, z.channel_freqs[2:8]))
    assert y.freq_align == 'center'

    y = z[:, 1:13]
    assert y.nchan == 12
    assert all(u.isclose(y.channel_freqs, z.channel_freqs[1:13]))
    assert y.freq_align == 'center'

    b = np.random.random((4, 2)) > 0.5
    y = z[2:8, :, b]
    assert y.nchan == z.nchan
    assert y.freq_align == z.freq_align
    assert all(u.isclose(y.channel_freqs, z.channel_freqs))
    assert u.isclose(y.center_freq, z.center_freq)

    y = z[2:8]
    assert y.nchan == z.nchan
    assert y.freq_align == z.freq_align
    assert all(u.isclose(y.channel_freqs, z.channel_freqs))
    assert u.isclose(y.center_freq, z.center_freq)


def test_radiosignal_slice_errors():
    x = np.ones((16, 16, 4, 2), dtype=np.float32)
    z = pb.RadioSignal(x, sample_rate=1*u.Hz, start_time=Time.now(),
                       center_freq=400*u.MHz, chan_bw=10*u.MHz,
                       freq_align='bottom')

    with pytest.raises(IndexError):
        _ = z[:, 'key']

    with pytest.raises(IndexError):
        _ = z[:, 4]

    with pytest.raises(IndexError):
        b = np.random.random(z.nchan) > 0.5
        _ = z[:, b]

    with pytest.raises(AssertionError):
        _ = z[:, 5:5]

    with pytest.raises(AssertionError):
        _ = z[:, 5:5]

    with pytest.raises(AssertionError):
        _ = z[:, ::2]

    with pytest.raises(AssertionError):
        _ = z[:, ::2]


def test_fullstokes_shape():
    for shape in [(10,), (4, 8), (4, 4, 2), (12, 0, 4)]:
        with pytest.raises(ValueError):
            x = np.empty(shape, dtype=np.float32)
            _ = pb.FullStokesSignal(x, sample_rate=1*u.Hz, center_freq=1*u.GHz,
                                    chan_bw=1*u.MHz)

    for shape in [(8, 4, 4), (4, 4, 4, 2)]:
        x = np.empty(shape, dtype=np.float32)
        _ = pb.FullStokesSignal(x, sample_rate=1*u.Hz, center_freq=1*u.GHz,
                                chan_bw=1*u.MHz)


def test_dualpol_shape():
    pol = 'linear'

    for shape in [(10,), (4, 8), (4, 4, 4), (12, 0, 2)]:
        x = np.empty(shape, dtype=np.complex128)
        with pytest.raises(ValueError):
            _ = pb.DualPolarizationSignal(x, sample_rate=1*u.Hz, pol_type=pol,
                                          center_freq=1*u.GHz)

    for shape in [(8, 4, 2), (4, 4, 2, 2)]:
        x = np.empty(shape, dtype=np.complex128)
        _ = pb.DualPolarizationSignal(x, sample_rate=1*u.Hz, pol_type=pol,
                                      center_freq=1*u.GHz)


def test_dualpol_pol_type():
    for pol in ['invalid', 3, ()]:
        x = np.empty((16, 4, 2), dtype=np.complex128)
        with pytest.raises(ValueError):
            _ = pb.DualPolarizationSignal(x, sample_rate=1*u.Hz, pol_type=pol,
                                          center_freq=1*u.GHz)

    for pol in ['linear', 'circular']:
        x = np.empty((16, 4, 2), dtype=np.complex128)
        z = pb.DualPolarizationSignal(x, sample_rate=1*u.Hz, pol_type=pol,
                                      center_freq=1*u.GHz)
        print(z)
        assert z.pol_type == pol



"""Reader module for pulsarbat."""

# flake8: noqa

from .base import AbstractReader
from .baseband_readers import *

try:
    from .dask_baseband_readers import *
except:  # pragma: no cover
    pass



"""Collection of handy utilities."""

import numpy as np
from scipy.fft import next_fast_len
# import astropy.units as u


__all__ = [
    'real_to_complex',
    'next_fast_len',
]


def real_to_complex(z, axis=0):
    """Convert a real baseband signal to a complex baseband signal.

    This function computes the analytic representation of the input
    signal via a Hilbert transform, throwing away negative frequency
    components. Then, the signal is shifted in frequency domain by -B/2
    where B is the bandwidth of the signal. Finally, the signal is
    decimated by a factor of 2, which results in the complex baseband
    representation of the input signal.

    If the input signal is complex-valued, only the real component is
    used.

    Parameters
    ----------
    z : `~numpy.ndarray`
        Input signal.
    axis : int, optional
        Axis over which to convert the signal. This will be the axis
        that represents time. Default is 0.

    Returns
    -------
    out : np.ndarray
        The complex baseband representation of the input signal.
    """
    if np.iscomplexobj(z):
        z = z.real

    out_dtype = np.complex128 if z.dtype == np.float64 else np.complex64

    # Pick the correct axis to work on
    if z.ndim > 1:
        ind = [np.newaxis] * z.ndim
        ind[axis] = slice(None)
    N = z.shape[axis]

    # Hilbert transform
    z = np.fft.fft(z, axis=axis)
    h = np.zeros(N)
    if N % 2 == 0:
        h[0] = h[N // 2] = 1
        h[1:N // 2] = 2
    else:
        h[0] = 1
        h[1:(N + 1) // 2] = 2
    if z.ndim > 1:
        h = h[tuple(ind)]
    z = np.fft.ifft(z * h, axis=axis)

    # Frequency shift signal by -B/2
    h = np.exp(-1j * np.pi / 2 * np.arange(N))
    if z.ndim > 1:
        h = h[tuple(ind)]
    z *= h

    # Decimate signal by factor of 2
    dec = [slice(None)] * z.ndim
    dec[axis] = slice(None, None, 2)
    z = z[tuple(dec)]
    return z.astype(out_dtype)


# def taper_function(freqs, bandwidth):
#     x = (freqs / bandwidth).to_value(u.one)
#     taper = 1 + (x / 0.48)**80
#     return taper



"""Dedispersion routines."""

import math
import numpy as np
import astropy.units as u
from ..core import RadioSignal, BasebandSignal

__all__ = [
    'DispersionMeasure', 'coherent_dedispersion', 'incoherent_dedispersion'
]


class DispersionMeasure(u.SpecificTypeQuantity):
    _equivalent_unit = _default_unit = u.pc / u.cm**3
    dispersion_constant = u.s * u.MHz**2 * u.cm**3 / u.pc / 2.41E-4

    def time_delay(self, f, ref_freq):
        """Time delay of frequencies relative to reference frequency."""
        coeff = self.dispersion_constant * self
        delay = coeff * (1 / f**2 - 1 / ref_freq**2)
        return delay.to(u.s)

    def sample_delay(self, f, ref_freq, sample_rate):
        """Sample delay of frequencies relative to reference frequency."""
        samples = self.time_delay(f, ref_freq) * sample_rate
        samples = samples.to_value(u.one)
        return samples

    def phase_delay(self, f, ref_freq):
        coeff = self.dispersion_constant * self
        phase = coeff * f * u.cycle * (1 / ref_freq - 1 / f)**2
        return phase.to_value(u.rad)

    def transfer_function(self, f, ref_freq):
        """Returns the transfer function for dedispersion."""
        transfer = np.exp(1j * self.phase_delay(f, ref_freq))
        return transfer.astype(np.complex64)


def coherent_dedispersion(z, DM, /, *, ref_freq, chirp=None):
    """Coherently dedisperses a baseband signal by a given dispersion measure.

    The output signal will be cropped on both ends to avoid wrap-around
    artifacts caused by dedispersion. This depends on how the reference
    frequency (`ref_freq`) compares to the band of the signal.

    The chirp function can be provided via the `chirp` argument which will be
    used instead of computing one from scratch. This can be useful in cases
    where the chirp function needs to be cached for efficiency.

    Parameters
    ----------
    z : `~pulsarbat.BasebandSignal`
        The signal to be transformed.
    DM : `~pulsarbat.DispersionMeasure`
        Dispersion measure by which to dedisperse `z`.
    ref_freq : `~astropy.units.Quantity`
        Reference frequency to dedisperse to.
    chirp: `~numpy.ndarray`, optional
        The dedispersion chirp function can be provided to avoid
        computing a new one. This is useful when the same dedispersion
        operation is being done on many blocks of the same size

    Returns
    -------
    out : `~pulsarbat.BasebandSignal`
        The dedispersed signal.
    """
    if not isinstance(z, BasebandSignal):
        raise TypeError("Signal must be a BasebandSignal object.")

    if chirp is None:
        f = z.channel_freqs[None] + np.fft.fftfreq(len(z), z.dt)[:, None]
        chirp = DM.transfer_function(f, ref_freq)

    x = np.fft.fft(z.data, axis=0)
    x = np.fft.ifft((x.T / chirp.T).T, axis=0)

    y = type(z).like(z, x)

    delay_top = DM.sample_delay(z.max_freq, ref_freq, z.sample_rate)
    delay_bot = DM.sample_delay(z.min_freq, ref_freq, z.sample_rate)

    start = math.ceil(-min(0, delay_top, delay_bot))
    stop = len(y) - math.ceil(+max(0, delay_top, delay_bot))

    return type(z).like(z, x)[start:stop]


def incoherent_dedispersion(z, DM, /, *, ref_freq):
    """Incoherently dedisperses a signal by a given dispersion measure.

    The output signal will be cropped on both ends to avoid wrap-around
    artifacts caused by dedispersion. This depends on how the reference
    frequency (`ref_freq`) compares to the band of the signal.

    Parameters
    ----------
    z : `~pulsarbat.RadioSignal`
        The signal to be transformed.
    DM : `~pulsarbat.DispersionMeasure`
        Dispersion measure by which to dedisperse `z`.
    ref_freq : `~astropy.units.Quantity`
        Reference frequency to dedisperse to.

    Returns
    -------
    out : `~pulsarbat.RadioSignal`
        The dedispersed signal.
    """
    if not isinstance(z, RadioSignal):
        raise TypeError("Signal must be a RadioSignal object.")

    delays = DM.sample_delay(z.channel_freqs, ref_freq, z.sample_rate)
    delays = delays.round().astype(int)

    crop_before = -min(0, delays[0], delays[-1])
    delays += crop_before
    N = len(z) - max(delays)

    x = np.stack([z.data[j:j+N, i] for i, j in enumerate(delays)], axis=1)

    if crop_before and z.start_time is not None:
        new_start = z.start_time + crop_before * z.dt
    else:
        new_start = z.start_time

    return type(z).like(z, x, start_time=new_start)



"""Tests for `pulsarbat.RadioSignal` and subclasses."""

import pytest
from pathlib import Path
from collections import namedtuple
import baseband
import numpy as np
import astropy.units as u
from astropy.time import Time
import pulsarbat as pb


Sample = namedtuple("Sample", ("name", "kwargs"))

DATA_DIR = Path(__file__).parent.absolute() / 'data'

SAMPLE_GUPPI = Sample(name=[DATA_DIR / f'fake.{i}.raw' for i in range(4)],
                      kwargs={'format': 'guppi'})
SAMPLE_DADA = Sample(name=str(DATA_DIR / 'sample.dada'),
                     kwargs={'format': 'dada'})
SAMPLE_STOKES_DADA = Sample(name=str(DATA_DIR / 'stokes_ef.dada'),
                            kwargs={'format': 'dada'})
SAMPLE_VDIF = Sample(name=str(DATA_DIR / 'sample.vdif'),
                     kwargs={'format': 'vdif'})


@pytest.mark.parametrize("reader", [pb.reader.DaskBasebandReader,
                                    pb.reader.BasebandReader])
@pytest.mark.parametrize("sample", [SAMPLE_GUPPI, SAMPLE_DADA, SAMPLE_VDIF,
                                    SAMPLE_STOKES_DADA])
def test_basebandreader(reader, sample):
    fh = baseband.open(sample.name, 'rs', **sample.kwargs)
    rdr = reader(sample.name, **sample.kwargs)

    assert u.isclose(fh.sample_rate, rdr.sample_rate)
    assert abs(fh.start_time - rdr.start_time) < 0.1 * u.ns
    assert abs(fh.stop_time - rdr.stop_time) < 0.1 * u.ns
    assert fh.sample_shape == rdr.sample_shape
    assert fh.dtype == rdr.dtype
    assert fh.shape[0] == len(rdr)

    x = fh.read(fh.shape[0])
    y = rdr.read(len(rdr))
    assert isinstance(y, pb.Signal)
    assert np.allclose(np.asarray(x), np.asarray(y))


def test_seek_and_tell():
    r = pb.reader.GUPPIRawReader(SAMPLE_GUPPI.name)
    N, dt = len(r), 320*u.ns

    for i in [0, N, N//2, N//3]:
        M = r.seek(i)
        assert r.tell() == M == i
        assert u.isclose(r.tell(unit=u.s), i*dt)
        assert abs(r.time - (r.start_time + i*dt)) < 0.1 * u.ns
        M = r.seek(i, whence=0)
        assert r.tell() == M == i
        M = r.seek(i*dt, whence='start')
        assert r.tell() == M == i

    for i in [-1, -10, -100, -1234]:
        M = r.seek(i, whence=2)
        assert r.tell() == N + i == M
        M = r.seek(i*dt, whence='end')
        assert r.tell() == N + i == M

    for i, j in [(100, -1), (200, 50), (4333, 0)]:
        r.seek(i)
        M = r.seek(j, whence=1)
        assert r.tell() == M == i + j
        r.seek(i*dt)
        M = r.seek(j*dt, whence='current')
        assert r.tell() == M == i + j

    t = r.start_time + 1000 * dt
    for whence in [0, 1, 2, 'start', 'current', 'end']:
        M = r.seek(t, whence=whence)
        assert r.tell() == M == 1000


def test_seek_errors():
    r = pb.reader.GUPPIRawReader(SAMPLE_GUPPI.name)
    bad_seeks = [-1, 40000, -1*u.s, -320*u.ns, 1*u.s,
                 Time('1997-07-11T12:34:55.99', format='isot'),
                 Time('1997-07-11T12:34:56.02', format='isot')]

    for i in bad_seeks:
        with pytest.raises(EOFError):
            r.seek(i)

    for whence in [3, 'curr', 'side']:
        with pytest.raises(ValueError):
            r.seek(10, whence=whence)


@pytest.mark.parametrize("reader", [pb.reader.BasebandRawReader,
                                    pb.reader.DaskBasebandRawReader])
@pytest.mark.parametrize("sideband", [True, False, np.arange(8) % 3 == 0])
def test_basebandrawreader(reader, sideband):
    r = reader(SAMPLE_VDIF.name, center_freq=0*u.Hz, sideband=sideband,
               **SAMPLE_VDIF.kwargs)
    assert len(r) == 20000
    _ = np.asarray(r.read(10000))
    assert r.tell() == 10000
    _ = np.asarray(r.read(9999))
    assert r.tell() == 19999
    _ = np.asarray(r.read(1))
    assert r.tell() == 20000
    with pytest.raises(EOFError):
        _ = r.read(1)


def test_basebandraw_errors():
    reader = pb.reader.BasebandRawReader

    for sideband in [1, (), np.arange(5) % 3 == 0]:
        with pytest.raises(ValueError):
            _ = reader(SAMPLE_VDIF.name, sideband=sideband,
                       center_freq=0*u.Hz, **SAMPLE_VDIF.kwargs)

    for fcen in [0, [12, 13]*u.Hz, 65*u.m]:
        with pytest.raises(ValueError):
            _ = reader(SAMPLE_VDIF.name, sideband=True,
                       center_freq=fcen, **SAMPLE_VDIF.kwargs)

    for falign in ['middle', 'sideways', 0]:
        with pytest.raises(ValueError):
            _ = reader(SAMPLE_VDIF.name, sideband=True, freq_align=falign,
                       center_freq=0*u.Hz, **SAMPLE_VDIF.kwargs)


@pytest.mark.parametrize("sb", [True, False, np.arange(8) % 3 == 0])
def test_basebandrawreader_consistency(sb):
    r1 = pb.reader.BasebandRawReader(SAMPLE_VDIF.name, center_freq=0*u.Hz,
                                     sideband=sb, **SAMPLE_VDIF.kwargs)

    r2 = pb.reader.DaskBasebandRawReader(SAMPLE_VDIF.name, center_freq=0*u.Hz,
                                         sideband=sb, **SAMPLE_VDIF.kwargs)

    x = r1.read(123)
    y = r2.read(123)
    assert np.allclose(np.asarray(x), np.asarray(y))

    x = r1.read(8192)
    y = r2.read(8192)
    assert np.allclose(np.asarray(x), np.asarray(y))


@pytest.mark.parametrize("reader", [pb.reader.GUPPIRawReader,
                                    pb.reader.DaskGUPPIRawReader])
def test_guppirawreader(reader):
    r = reader(SAMPLE_GUPPI.name)
    assert u.isclose(r.sample_rate, 3.125*u.MHz)
    assert u.isclose(r.center_freq, 344.1875*u.MHz)
    assert r.pol_type == 'linear'
    assert r.dtype == np.complex64
    assert r.freq_align == 'center'
    assert r.sideband is True

    st = Time('1997-07-11T12:34:56', format='isot', precision=9)
    z = r.read(1)
    assert abs(z.start_time - st) < 0.1 * u.ns
    assert u.isclose(z.bandwidth, 12.5*u.MHz)
    fs = [339.5, 342.625, 345.75, 348.875] * u.MHz
    assert u.allclose(fs, z.channel_freqs)
    y = np.array([[5-5j, -31-14j], [-17+6j, -21+0j], [12-22j, 2-12j],
                  [15+17j, -5+21j]], dtype=np.complex64)
    assert np.allclose(np.array(z), y)


@pytest.mark.parametrize("reader", [pb.reader.DADAStokesReader,
                                    pb.reader.DaskDADAStokesReader])
def test_dadastokesreader(reader):
    r = reader(SAMPLE_STOKES_DADA.name)
    assert u.isclose(r.sample_rate, 1 / (131.072 * u.us))
    assert u.isclose(r.center_freq, 7*u.GHz)
    assert r.dtype == np.float32
    assert r.freq_align == 'top'
    assert u.isclose(r.chan_bw, (2 * u.GHz) / 2048)

    st = Time('2019-01-13T15:57:41', format='isot', precision=9)
    z = r.read(1)
    assert abs(z.start_time - st) < 0.1 * u.ns
    assert u.isclose(z.bandwidth, 2*u.GHz)
    assert u.isclose(z.channel_freqs[-1], 8*u.GHz)
    assert u.isclose(z.channel_freqs[1023], z.center_freq)
    y = np.array([[[21, 1, -3, -1], [22, 2, -2, -1], [24, 4, -3, 0],
                   [22, 3, 3, 0]]], dtype=np.float32)
    assert np.allclose(np.array(z[:, 500:504]), y)


def test_dadastokesreader_error():
    with pytest.raises(ValueError):
        _ = pb.reader.DADAStokesReader(SAMPLE_DADA.name)



"""Tests for `pulsarbat.Signal` class."""

import pytest
import numpy as np
import dask.array as da
import astropy.units as u
from astropy.time import Time
import pulsarbat as pb


@pytest.mark.parametrize("fn", [np.ones, da.ones])
@pytest.mark.parametrize("ts", [None, Time.now()])
@pytest.mark.parametrize("shape", [(16,), (8, 4)])
@pytest.mark.parametrize("sample_rate", [1 * u.Hz, 33 * u.kHz])
@pytest.mark.parametrize("dtype", [np.float32, np.complex64])
def test_signal_basic(fn, ts, shape, sample_rate, dtype):
    x = fn(shape, dtype=dtype)
    z = pb.Signal(x, sample_rate=sample_rate, start_time=ts)

    assert isinstance(z, pb.Signal)
    print(z)
    repr(z)

    assert z.shape == shape
    assert z.ndim == len(shape)
    assert len(z) == shape[0]
    assert z.sample_shape == shape[1:]
    assert z.dtype == x.dtype
    assert type(z.data) is type(x)
    assert np.all(np.array(z) == np.array(x))

    assert z.sample_rate == sample_rate
    assert z.sample_rate is not sample_rate
    assert u.isclose(z.dt, 1 / sample_rate)
    assert u.isclose(z.time_length, shape[0] / sample_rate)

    if ts is None:
        assert z.start_time is None
        assert z.stop_time is None
        assert Time.now() not in z
    else:
        assert z.start_time == ts
        assert z.stop_time == ts + (shape[0] / sample_rate)
        assert z.start_time in z
        assert z.stop_time not in z
        assert ts + 1 * u.ns in z


@pytest.mark.parametrize("dt", [0 * u.s, np.sqrt(2) * u.day, 1/(4/7 * u.Hz)])
def test_time_contains(dt):
    st = Time.now()
    x = np.ones((16, 2), dtype=np.float32)
    z = pb.Signal(x, sample_rate=1*u.Hz, start_time=st)
    N = 10
    ts = ((st + np.arange(-N, len(z)+N) * u.s) + dt) - dt
    check = np.array([t in z for t in ts])
    assert np.all(check[N:-N])
    assert not np.all(check[:N])
    assert not np.all(check[-N:])


@pytest.mark.parametrize("shape", [(), (0, ), (0, 4, 2)])
def test_shape_errors(shape):
    with pytest.raises(ValueError):
        _ = pb.Signal(np.empty(shape), sample_rate=1 * u.Hz)


def test_sample_rate_errors():
    for x in [99., 4*u.m, [5, 6]*u.Hz, -1*u.Hz]:
        with pytest.raises(ValueError):
            _ = pb.Signal(np.empty((8, 4)), sample_rate=x)


@pytest.mark.parametrize("ts", [Time([59843, 45678, 65678], format='mjd'),
                                59867.2442234])
def test_start_time_errors(ts):
    with pytest.raises(ValueError):
        _ = pb.Signal(np.empty((8, 4)), sample_rate=1 * u.Hz, start_time=ts)


class ArbitrarySignal(pb.Signal):
    _req_dtype = (np.int8, np.int16)
    _req_shape = (7, None, 4, None)

    def __init__(self, z, /, *, sample_rate, foo):
        self._foo = foo
        super().__init__(z, sample_rate=sample_rate)

    @property
    def foo(self):
        return self._foo


@pytest.mark.parametrize("dtype", [np.int8, np.int16])
def test_dtype_constraints(dtype):
    x = np.random.uniform(-50, 50, (7, 2, 4, 1)).astype(dtype)
    z = ArbitrarySignal(x, sample_rate=1*u.Hz, foo='bar')
    assert z.dtype == dtype


@pytest.mark.parametrize("dtype", [np.float32, np.int32])
def test_dtype_constraints_error(dtype):
    x = np.random.uniform(-50, 50, (7, 2, 4, 1)).astype(dtype)
    with pytest.raises(ValueError):
        _ = ArbitrarySignal(x, sample_rate=1*u.Hz, foo='bar')


@pytest.mark.parametrize("shape", [(7,), (7, 2), (7, 1, 4), (7, 2, 3, 2)])
def test_shape_constraints(shape):
    x = np.random.uniform(-50, 50, shape).astype(np.int8)
    with pytest.raises(ValueError):
        _ = ArbitrarySignal(x, sample_rate=1*u.Hz, foo='bar')


def test_signal_like():
    x1 = np.random.uniform(-50, 50, (7, 1, 4, 1)).astype(np.int8)
    x2 = np.random.uniform(-50, 50, (7, 1, 4, 1)).astype(np.int8)

    z1 = ArbitrarySignal(x1, sample_rate=1*u.Hz, foo='bar')
    z2 = pb.Signal.like(z1)

    assert np.array_equal(np.array(z1), np.array(z2))
    assert z1.sample_rate == z2.sample_rate
    assert z1.start_time == z2.start_time
    assert type(z2) is pb.Signal

    z2 = pb.Signal.like(z1, x2)
    assert not np.array_equal(np.array(z1), np.array(z2))

    z2 = pb.Signal.like(z1, sample_rate=2*z1.sample_rate)
    assert u.isclose(z2.sample_rate, 2 * z1.sample_rate)

    with pytest.raises(ValueError):
        _ = ArbitrarySignal.like(z2)

    z3 = ArbitrarySignal.like(z2, foo='bar')
    assert np.array_equal(np.array(z1), np.array(z3))
    assert z1.foo == z3.foo == 'bar'
    assert u.isclose(z3.sample_rate, 2 * z1.sample_rate)


def test_signal_slice():
    x = np.ones((32, 4, 2), dtype=np.float32)
    z = pb.Signal(x, sample_rate=1*u.Hz, start_time=Time.now())

    y = z[2:8]
    assert len(y) == 6
    assert np.abs(z.start_time - (y.start_time - 2*u.s)) < 0.1 * u.ns
    assert u.isclose(z.sample_rate, y.sample_rate)
    assert u.isclose(y.time_length, 6 * u.s)

    y = z[::4]
    assert len(y) == 8
    assert np.abs(z.start_time - y.start_time) < 0.1 * u.ns
    assert u.isclose(y.sample_rate, 0.25 * u.Hz)
    assert u.isclose(z.time_length, y.time_length)

    y = z[:, [0, 2]]
    assert len(y) == len(z)
    assert np.abs(z.start_time - y.start_time) < 0.1 * u.ns
    assert u.isclose(z.sample_rate, y.sample_rate)
    assert u.isclose(z.time_length, y.time_length)

    b = np.random.random((4, 2)) > 0.5
    y = z[:, b]
    assert len(y) == len(z)
    assert np.abs(z.start_time - y.start_time) < 0.1 * u.ns
    assert u.isclose(z.sample_rate, y.sample_rate)
    assert u.isclose(z.time_length, y.time_length)

    z = pb.Signal(x, sample_rate=1*u.Hz)
    assert z[2:].start_time is None
    assert z[:6].start_time is None
    assert z[::4].start_time is None


def test_signal_slice_errors():
    x = np.ones((32, 4, 2), dtype=np.float32)
    st = Time.now()
    z = pb.Signal(x, sample_rate=1*u.Hz, start_time=st)

    with pytest.raises(IndexError):
        _ = z['key']

    with pytest.raises(IndexError):
        _ = z[4]

    with pytest.raises(IndexError):
        b = np.random.random(len(z)) > 0.5
        _ = z[b]

    with pytest.raises(AssertionError):
        _ = z[::-2]

    with pytest.raises(AssertionError):
        _ = z[5:5]

    with pytest.raises(AssertionError):
        _ = z[10:5]



"""Signal functions."""

# flake8: noqa

from . import core
from . import dedispersion

from .core import *
from .dedispersion import *

__all__ = []
__all__.extend(core.__all__)
__all__.extend(dedispersion.__all__)



"""Core signal functions."""

import numpy as np
import astropy.units as u
from ..core import Signal, RadioSignal

__all__ = ['concatenate', ]


def concatenate(signals, axis=0):
    """Concatenates multiple signals in time or frequency.

    All signals must have the same type. Signals must also have the same
    `sample_rate` when concatenating along time or the same `chan_bw`
    when concatenating along frequency. The concatenated signal will
    inherit all other attributes from the first signal provided
    (`signals[0]`) except for `center_freq` when concatenating along
    frequency, of course (there is no problem when concatenating along
    time as the `start_time` is used as a reference which is naturally
    extracted from `signals[0]`).

    Parameters
    ----------
    signals : sequence of `~Signal`-like
        Ordered sequence of signals to concatenate.
    axis : 0 or 1, optional
        Axis along which to concatenate signals. `0` (default)
        indicates concatenation along time, `1` indicates concatenation
        along frequency.

    Returns
    -------
    out : `~Signal`-like
        Concatenated signal of same type as input signals.
    """
    def times_are_equal(t0, t1):
        if t0 is None:
            return t1 is None
        else:
            return abs(t1 - t0) < 0.1 * u.ns

    if axis not in (0, 1):
        raise ValueError("Invalid axis. Must be either 0 or 1.")

    if not isinstance(signals[0], {0: Signal, 1: RadioSignal}[axis]):
        err = {0: "Must be Signal object when axis=0.",
               1: "Must be RadioSignal object when axis=1."}
        raise TypeError(err[axis])

    _type = type(signals[0])
    if not all(type(s) is _type for s in signals):
        raise TypeError("All signals not of same type!")

    _kwargs = dict()
    if axis == 0:
        ref = signals[0].sample_rate
        if not all(u.isclose(ref, s.sample_rate) for s in signals):
            raise ValueError("Signals don't have the same sample_rate!")

        if not all(times_are_equal(i.stop_time, j.start_time)
                   for i, j in zip(signals, signals[1:])):
            raise ValueError("Signals not contiguous in time.")
    else:
        ref = signals[0].chan_bw
        if not all(u.isclose(ref, s.chan_bw) for s in signals):
            raise ValueError("Signals don't have the same chan_bw!")

        if not all(u.isclose(j.channel_freqs[0] - i.channel_freqs[-1], ref)
                   for i, j in zip(signals, signals[1:])):
            raise ValueError("Signals not contiguous in frequency.")

        f0, f1 = signals[0].channel_freqs[0], signals[-1].channel_freqs[-1]
        _kwargs['center_freq'] = (f0 + f1) / 2
        _kwargs['freq_align'] = 'center'

    z = np.concatenate([s.data for s in signals], axis=axis)
    return _type.like(signals[0], z, **_kwargs)



"""Tests for `pulsarbat.RadioSignal` and subclasses."""

import pytest
import numpy as np
import dask.array as da
import astropy.units as u
import pulsarbat as pb

SHAPE = (4096, 8, 2)


@pytest.fixture(params=[True, False])
def use_dask(request):
    return request.param


@pytest.fixture
def complex_noise(use_dask):
    f = da.random.standard_normal if use_dask else np.random.standard_normal
    return (f(SHAPE) + 1j * f(SHAPE)).astype(np.complex128)


@pytest.mark.parametrize("pol_type", ['linear', 'circular'])
def test_pol_reversibility(complex_noise, pol_type):
    z = pb.DualPolarizationSignal(complex_noise, pol_type=pol_type,
                                  sample_rate=1*u.MHz, center_freq=1*u.GHz)

    if pol_type == 'linear':
        x = z.to_linear()
        y = z.to_circular().to_linear()
    else:
        x = z.to_circular()
        y = z.to_linear().to_circular()

    assert z.pol_type == x.pol_type
    assert z.pol_type == y.pol_type

    x, y, z = np.array(x), np.array(y), np.array(z)
    assert np.allclose(z.real, x.real)
    assert np.allclose(z.real, y.real)
    assert np.allclose(z.imag, x.imag)
    assert np.allclose(z.imag, y.imag)


@pytest.mark.parametrize("pol_type", ['linear', 'circular'])
def test_to_stokes(complex_noise, pol_type):
    z = pb.DualPolarizationSignal(complex_noise, pol_type=pol_type,
                                  sample_rate=1*u.MHz, center_freq=1*u.GHz)
    x = z.to_linear().to_stokes()
    y = z.to_circular().to_stokes()
    x, y = np.array(x), np.array(y)
    assert np.allclose(x.real, y.real)
    assert np.allclose(x.imag, y.imag)



"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'astropy>=4.1',
    'numpy>=1.19',
    'scipy>=1.5',
    'baseband>=4.0.2'
]

setup(
    author="Nikhil Mahajan",
    author_email='mahajan@astro.utoronto.ca',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
    ],
    description=(
        'pulsarbat (PULSAR Baseband Analysis Tools) is a Python package '
        'for analysis of radio baseband observations of pulsars.'),
    install_requires=requirements,
    python_requires='>=3.8',
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='pulsarbat',
    name='pulsarbat',
    packages=find_packages(include=['pulsarbat', 'pulsarbat.*']),
    url='https://github.com/theXYZT/pulsarbat',
    version='0.0.4',
    zip_safe=False,
)



"""Base module for reader classes."""

import operator
from abc import ABCMeta, abstractmethod
import astropy.units as u
from astropy.time import Time
from ..core import Signal

__all__ = ['AbstractReader', 'OutOfBoundsError']


class OutOfBoundsError(EOFError):
    """Raised when stream position is out of bounds."""
    pass


class AbstractReader(metaclass=ABCMeta):
    """Abstract base class for readers.

    Readers are used to read data into `~Signal` objects. Subclasses
    must necessarily define:
      * `__len__()` -- Returns length of stream in number of samples.
      * `sample_rate` -- property, number of samples per unit time.
      * `start_time` -- property, timestamp at start of stream.
      * `_read_stream(n)` -- Reads `n` samples and returns an array-like.
    """
    def __init__(self):
        self.offset = 0

    @abstractmethod
    def __len__(self):
        """Length of stream in number of samples."""
        pass  # pragma: no cover

    @property
    @abstractmethod
    def sample_rate(self):
        """Sample rate (number of samples per unit time)."""
        pass  # pragma: no cover

    @property
    @abstractmethod
    def start_time(self):
        """Timestamp at first sample."""
        pass  # pragma: no cover

    @property
    def stop_time(self):
        """Timestamp after last sample."""
        stop = self.start_time + (len(self) / self.sample_rate)
        return Time(stop, format='isot', precision=9)

    @property
    def time(self):
        """Timestamp at current read position."""
        curr = self.start_time + (self.tell() / self.sample_rate)
        return Time(curr, format='isot', precision=9)

    def seek(self, offset, whence=0):
        """Change read position to the given offset.

        Offset is interpreted relative to position indicated by `whence`:
          * `0` or `'start'` -- Start of the stream (default). `offset`
            must be positive.
          * `1` or `'current'` -- Current stream position. `offset` must
            be positive or negative.
          * `2` or `'end'` -- End of the stream. `offset` must be negative.

        Parameters
        ----------
        offset : int, `~astropy.units.Quantity`, or `~astropy.time.Time`
            Offset to move to. Can be an `int` (number of samples),
            an astropy Quantity in time units, or an absolute time as an
            astropy Time object (`whence` has no effect in this case).
            For the latter two, the seek position is moved to the
            nearest integer sample.
        whence : {0, 1, 2, 'start', 'current', or 'end'}, optional
            Position that `offset` is taken relative to. Default is `0`,
            which is the start of the stream.

        Returns
        -------
        offset : int
            The new absolute position (in number of samples).
        """
        try:
            offset = operator.index(offset)
        except Exception:
            try:
                offset = offset - self.start_time
            except Exception:
                pass
            else:
                whence = 0

            offset = int((offset * self.sample_rate).to(u.one).round())

        if whence in {0, 'start'}:
            self.offset = offset
        elif whence in {1, 'current'}:
            self.offset = self.offset + offset
        elif whence in {2, 'end'}:
            self.offset = len(self) + offset
        else:
            raise ValueError("Invalid 'whence'. Should be 0 or 'start', 1 or "
                             "'current', or 2 or 'end'.")

        if self.offset < 0 or self.offset > len(self):
            raise OutOfBoundsError("Cannot seek beyond bounds of stream!")
        return self.offset

    def tell(self, unit=None):
        """Return current stream position.

        Parameters
        ----------
        unit : `~astropy.units.Unit`, optional
            Time unit the offset should be returned in. By default, the
            current stream position is returned as an integer.

        Returns
        -------
        offset : int or `~astropy.units.Quantity`
            Current stream position.
        """
        if unit is None:
            return self.offset

        return (self.offset / self.sample_rate).to(unit)

    @abstractmethod
    def _read_stream(self, n, /, **kwargs):
        """Read N samples from current stream position into array-like."""
        pass  # pragma: no cover

    def _to_signal(self, z, /, start_time):
        """Return Signal containing given data."""
        return Signal(z, sample_rate=self.sample_rate, start_time=start_time)

    def read(self, n, /, **kwargs):
        """Read N samples from current stream position into Signal."""
        curr_time = self.time

        if self.tell() + n > len(self):
            raise OutOfBoundsError("Cannot read beyond end of stream")

        z = self._read_stream(n, **kwargs)
        _ = self.seek(n, whence=1)
        return self._to_signal(z, curr_time)
